"""Yeast database query functions."""
from intermine.webservice import Service
import coral
# TODO: Use httplib instead if we only need to do one requests-style function
try:
    import requests

except ImportError:
    print "requests module could not be imported, fetching sequences will " + \
          "not work."


def fetch_yeast_locus_sequence(locus_name, flanking_size=0):
    """Acquire a sequence from SGD http://www.yeastgenome.org.

    :param locus_name: Common name or systematic name for the locus (e.g. ACT1
                       or YFL039C).
    :type locus_name: str
    :param flanking_size: The length of flanking DNA (on each side) to return
    :type flanking_size: int

    """
    service = Service("http://yeastmine.yeastgenome.org/yeastmine/service")

    # Get a new query on the class (table) you will be querying:
    query = service.new_query("Gene")

    if flanking_size > 0:

        # The view specifies the output columns
        # secondaryIdentifier: the systematic name (e.g. YFL039C)
        # symbol: short name (e.g. ACT1)
        # length: sequence length
        # flankingRegions.direction: Upstream or downstream (or both) of locus
        # flankingRegions.sequence.length: length of the flanking regions
        # flankingRegions.sequence.residues: sequence of the flanking regions
        query.add_view("secondaryIdentifier", "symbol", "length",
                       "flankingRegions.direction",
                       "flankingRegions.sequence.length",
                       "flankingRegions.sequence.residues")

        # You can edit the constraint values below
        query.add_constraint("flankingRegions.direction", "=", "both",
                             code="A")
        query.add_constraint("Gene", "LOOKUP", locus_name, "S. cerevisiae",
                             code="B")
        query.add_constraint("flankingRegions.distance", "=",
                             "{:.1f}kb".format(flanking_size / 1000.),
                             code="C")
        # Uncomment and edit the code below to specify your own custom logic:
        query.set_logic("A and B and C")

        # TODO: What to do when there"s more than one result?
        first_result = query.rows().next()
        # FIXME: Use logger module instead
        # print first_result["secondaryIdentifier"]
        # print first_result["symbol"], row["length"]
        # print first_result["flankingRegions.direction"]
        # print first_result["flankingRegions.sequence.length"]
        # print first_result["flankingRegions.sequence.residues"]

        seq = coral.DNA(first_result["flankingRegions.sequence.residues"])
        # TODO: add more metadata

    elif flanking_size == 0:
        # The view specifies the output columns
        query.add_view("primaryIdentifier", "secondaryIdentifier", "symbol",
                       "name", "sgdAlias", "organism.shortName",
                       "sequence.length", "sequence.residues", "description",
                       "qualifier")

        query.add_constraint("status", "IS NULL", code="D")
        query.add_constraint("status", "=", "Active", code="C")
        query.add_constraint("qualifier", "IS NULL", code="B")
        query.add_constraint("qualifier", "!=", "Dubious", code="A")
        query.add_constraint("Gene", "LOOKUP", locus_name, "S. cerevisiae",
                             code="E")

        # Your custom constraint logic is specified with the code below:
        query.set_logic("(A or B) and (C or D) and E")

        first_result = query.rows().next()
        seq = coral.DNA(first_result["sequence.residues"])
    else:
        print "Problem with the flanking region size...."
        seq = coral.DNA("")

    return seq


def get_yeast_sequence(chromosome, start, end, reverse_complement=False):
    """Acquire a sequence from SGD http://www.yeastgenome.org
    :param chromosome: Yeast chromosome.
    :type chromosome: int
    :param start: A biostart.
    :type start: int
    :param end: A bioend.
    :type end: int
    :param reverse_complement: Get the reverse complement.
    :type revervse_complement: bool
    :returns: A DNA sequence.
    :rtype: coral.DNA

    """
    if start != end:
        if reverse_complement:
            rev_option = "-REV"
        else:
            rev_option = ""
        param_url = "&chr=" + str(chromosome) + "&beg=" + str(start) + \
                    "&end=" + str(end) + "&rev=" + rev_option
        url = "http://www.yeastgenome.org/cgi-bin/getSeq?map=a2map" + \
            param_url

        res = requests.get(url)
        # ok... sadely, I contacted SGD and they haven;t implemented this so
        # I have to parse their yeastgenome page, but
        # it is easy between the raw sequence is between <pre> tags!

        # warning that"s for the first < so we need +5!
        begin_index = res.text.index("<pre>")

        end_index = res.text.index("</pre>")
        sequence = res.text[begin_index + 5:end_index]
        sequence = sequence.replace("\n", "").replace("\r", "")
    else:
        sequence = ""

    return coral.DNA(sequence)


def get_yeast_gene_location(gene_name):
    """Acquire the location of a gene from SGD http://www.yeastgenome.org
    :param gene_name: Name of the gene.
    :type gene_name: string
    :returns location: [int: chromosome, int:biostart, int:bioend, int:strand]
    :rtype location: list

    """
    service = Service("http://yeastmine.yeastgenome.org/yeastmine/service")

    # Get a new query on the class (table) you will be querying:
    query = service.new_query("Gene")

    # The view specifies the output columns
    query.add_view("primaryIdentifier", "secondaryIdentifier", "symbol",
                   "name", "organism.shortName",
                   "chromosome.primaryIdentifier",
                   "chromosomeLocation.start", "chromosomeLocation.end",
                   "chromosomeLocation.strand")

    # Uncomment and edit the line below (the default) to select a custom sort
    # order:
    # query.add_sort_order("Gene.primaryIdentifier", "ASC")

    # You can edit the constraint values below
    query.add_constraint("organism.shortName", "=", "S. cerevisiae",
                         code="B")
    query.add_constraint("Gene", "LOOKUP", gene_name, code="A")

    # Uncomment and edit the code below to specify your own custom logic:
    # query.set_logic("A and B")
    chromosomes = {"chrI": 1,
                   "chrII": 2,
                   "chrIII": 3,
                   "chrIV": 4,
                   "chrV": 5,
                   "chrVI": 6,
                   "chrVII": 7,
                   "chrVIII": 8,
                   "chrIX": 9,
                   "chrX": 10,
                   "chrXI": 11,
                   "chrXII": 12,
                   "chrXIII": 13,
                   "chrXIV": 14,
                   "chrXV": 15,
                   "chrXVI": 16}
    first_result = query.rows().next()

    return [chromosomes[first_result["chromosome.primaryIdentifier"]],
            first_result["chromosomeLocation.start"],
            first_result["chromosomeLocation.end"],
            int(first_result["chromosomeLocation.strand"])]


def get_gene_id(gene_name):
    """Retrieve systematic yeast gene name from the common name.

    :param gene_name: Common name for yeast gene (e.g. ADE2).
    :type gene_name: str
    :returns: Systematic name for yeast gene (e.g. YOR128C).
    :rtype: str

    """
    service = Service("http://yeastmine.yeastgenome.org/yeastmine/service")

    # Get a new query on the class (table) you will be querying:
    query = service.new_query("Gene")

    # The view specifies the output columns
    query.add_view("primaryIdentifier", "secondaryIdentifier", "symbol",
                   "name", "sgdAlias", "crossReferences.identifier",
                   "crossReferences.source.name")

    # Uncomment and edit the line below (the default) to select a custom sort
    # order:
    # query.add_sort_order("Gene.primaryIdentifier", "ASC")

    # You can edit the constraint values below
    query.add_constraint("organism.shortName", "=", "S. cerevisiae", code="B")
    query.add_constraint("Gene", "LOOKUP", gene_name, code="A")

    # Uncomment and edit the code below to specify your own custom logic:
    # query.set_logic("A and B")

    for row in query.rows():
        gid = row["secondaryIdentifier"]
    return gid


def get_yeast_promoter_ypa(gene_name):
    """Retrieve promoter from Yeast Promoter Atlas
    (http://ypa.csbb.ntu.edu.tw).

    :param gene_name: Common name for yeast gene.
    :type gene_name: str
    :returns: Double-stranded DNA sequence of the promoter.
    :rtype: coral.DNA

    """
    loc = get_yeast_gene_location(gene_name)
    gid = get_gene_id(gene_name)
    ypa_baseurl = "http://ypa.csbb.ntu.edu.tw/do"
    params = {"act": "download",
              "nucle": "InVitro",
              "right": str(loc[2]),
              "left": str(loc[1]),
              "gene": str(gid),
              "chr": str(loc[0])}

    response = requests.get(ypa_baseurl, params=params)
    text = response.text
    # FASTA records are just name-sequence pairs split up by > e.g.
    # >my_dna_name
    # GACGATA
    # TODO: most of this is redundant, as we just want the 2nd record
    record_split = text.split(">")
    record_split.pop(0)
    parsed = []
    for record in record_split:
        parts = record.split("\n")
        sequence = coral.DNA("".join(parts[1:]))
        sequence.name = parts[0]
        parsed.append(sequence)

    return parsed[1]



'''PCR reaction(s).'''


class AmbiguousPrimingError(Exception):
    """Primer binds to more than one place on a template."""


class PrimerBindError(Exception):
    """Primer did not bind correctly."""


def pcr(template, primer1, primer2):
    '''Simulate a PCR (no support for ambiguous PCRs).

    :param template: DNA template from which to PCR.
    :type template: coral.DNA
    :param primer1: First PCR primer.
    :type primer1: coral.Primer
    :param primer2: First PCR primer.
    :type primer2: coral.Primer
    :returns: A dsDNA Amplicon.
    :rtype: coral.DNA
    :raises: Exception if a primer binds more than once on the template.
             Exception if primers bind in overlapping sequence of the template.
             Exception if the PCR would work on a circular version of the
             template (implies that input was linear).

    '''
    # FIXME: using the wrong primers/template produces a useless error.
    # make the error useful!
    # Find match in top or bottom strands for each primer
    p1_matches = template.locate(primer1.anneal)
    p2_matches = template.locate(primer2.anneal)

    # HEY FIX THIS - should find an ambiguity
    # Make sure there's no ambiguities
    p1_bind = sum([len(match) for match in p1_matches])
    p2_bind = sum([len(match) for match in p2_matches])

    def msg(location):
        return "Top strand: {}, Bottom strand: {}".format(location[0],
                                                          location[1])

    if p1_bind > 1:
        raise AmbiguousPrimingError("Primer 1, {}".format(msg(p1_matches)))
    if p2_bind > 1:
        raise AmbiguousPrimingError("Primer 2, {}".format(msg(p2_matches)))
    if not p1_bind and not p2_bind:
        raise PrimerBindError("Neither primer binds the template")
    if not p1_bind:
        raise PrimerBindError("Primer 1 does not bind the template")
    if not p2_bind:
        raise PrimerBindError("Primer 2 does not bind the template")

    # Make 'reverse' index useful for slicing
    fwds = p1_matches[0] + p2_matches[0]
    revs = p2_matches[1] + p1_matches[1]
    fwd = fwds[0]
    rev = len(template) - revs[0]
    # TODO: circular search will muck things up. If primers are at the very
    # beginning or end of the plasmid coordinates, things will get weird
    # TODO: Should actually just evaluate the length of the product prior
    # to adding primer overhangs, compare to length of anneal regions.
    #   But would need to keep track of sign - fwd - rev != rev - fwd
    '''Notes on PCR amplification decisions:
        If rev == fwd, primers should ampify entire plasmid
        If rev - fwd >= max(len(rev), len(fwd)), amplify sequence normally
        If rev - fwd < max(len(rev), len(fwd)), raise exception - who knows
        how this construct will amplify

    '''
    if rev - fwd < max(len(primer1), len(primer2)) and rev - fwd > 0:
        raise Exception('Primers bind in one another, no solution implemented')

    # Subset
    if rev > fwd:
        amplicon = template[fwd:rev]
    else:
        if template.topology == 'circular':
            amplicon = template[fwd:] + template[:rev]
        else:
            raise Exception('Primers would amplify if template were circular.')
    if primer1.overhang:
        amplicon = primer1.overhang.to_ds() + amplicon
    if primer2.overhang:
        amplicon += primer2.overhang.to_ds().reverse_complement()
    return amplicon



'''DNA object classes.'''
import collections
import os
import re
import shutil
import subprocess
import tempfile
import coral.analysis
import coral.reaction
import coral.seqio
from ._sequence import process_seq, reverse_complement
from ._sequence import NucleotideSequence


class DNA(NucleotideSequence):
    '''DNA sequence.'''
    def __init__(self, dna, bottom=None, topology='linear', stranded='ds',
                 features=None, run_checks=True, id=None, name=''):
        '''
        :param dna: Input sequence (DNA).
        :type dna: str
        :param bottom: Manual input of bottom-strand sequence. Enables both
                       mismatches and initializing ssDNA.
        :type bottom: str
        :param topology: Topology of DNA - 'linear' or 'circular'.
        :type topology: str
        :param stranded: Strandedness of DNA - 'ss' for single-stranded or
                         'ds' for double-stranded.
        :type stranded: str
        :param features: List of annotated features.
        :type features: list
        :param run_checks: Check inputs / formats (disabling increases speed):
                           alphabet check
                           case
        :type run_checks: bool
        :param id: An optional (unique) id field for your DNA sequence.
        :type id: str
        :param name: Optional name field for your DNA sequence.
        :type name: str
        :returns: coral.DNA instance.
        :rtype: coral.DNA
        :raises: ValueError if an element of `features` isn't of type
                 coral.Feature.
                 ValueError if top and bottom strands have different lengths.
                 ValueError if top and bottom strands are not complementary.

        '''
        # Convert to uppercase, run alphabet check
        super(DNA, self).__init__(dna, 'dna', features=features,
                                  run_checks=run_checks)
        # Set topology
        self.topology = topology
        # Set strandedness
        self.stranded = stranded
        # If bottom was specified, check it + add it
        if bottom:
            self._bottom = bottom
            if run_checks:
                self._bottom = process_seq(bottom, 'dna')
                if len(self._bottom) != len(self._sequence):
                    msg = 'Top and bottom strands are difference lengths.'
                    raise ValueError(msg)
        else:
            self._bottom = ''.join(['-' for x in self._sequence])
            # NOTE: inefficient to assign blanks the rev comp, but cleaner code
            if stranded == 'ds':
                self._bottom = str(reverse_complement(self._sequence, 'dna'))
        # Set id
        self.id = id
        # Set name
        self.name = name

    def ape(self, ape_path=None):
        '''Open in ApE.'''
        cmd = 'ApE'
        if ape_path is None:
            # Check for ApE in PATH
            ape_executables = []
            for path in os.environ['PATH'].split(os.pathsep):
                exepath = os.path.join(path, cmd)
                ape_executables.append(os.access(exepath, os.X_OK))
            if not any(ape_executables):
                raise Exception('Ape not in PATH. Use ape_path kwarg.')
        else:
            cmd = ape_path
        # Check whether ApE exists in PATH
        tmp = tempfile.mkdtemp()
        if self.name is not None and self.name:
            filename = os.path.join(tmp, '{}.ape'.format(self.name))
        else:
            filename = os.path.join(tmp, 'tmp.ape')
        coral.seqio.write_dna(self, filename)
        process = subprocess.Popen([cmd, filename])
        # Block until window is closed
        try:
            process.wait()
            shutil.rmtree(tmp)
        except KeyboardInterrupt:
            shutil.rmtree(tmp)

    def bottom(self):
        '''Return the raw string of the Crick (bottom) strand.

        :returns: The Crick strand.
        :rtype: str

        '''
        return self._bottom

    def copy(self):
        '''Create a copy of the current instance.

        :returns: A safely-editable copy of the current sequence.
        :rtype: coral.DNA

        '''
        # Significant performance improvements by skipping alphabet check
        features_copy = [feature.copy() for feature in self.features]
        return type(self)(self._sequence, bottom=self._bottom,
                          topology=self.topology, stranded=self.stranded,
                          features=features_copy, id=self.id, name=self.name,
                          run_checks=False)

    def circularize(self):
        '''Circularize linear DNA.

        :returns: A circularized version of the current sequence.
        :rtype: coral.DNA

        '''
        if self.top()[-1] == '-' and self.bottom()[0] == '-':
            raise ValueError('Cannot circularize - termini disconnected.')
        if self.bottom()[-1] == '-' and self.top()[0] == '-':
            raise ValueError('Cannot circularize - termini disconnected.')

        copy = self.copy()
        copy.topology = 'circular'
        return copy

    def extract(self, name, remove_subfeatures=False):
        return super(DNA, self).extract(name, 'N',
                                        remove_subfeatures=remove_subfeatures)

    def flip(self):
        '''Flip the DNA - swap the top and bottom strands.

        :returns: Flipped DNA (bottom strand is now top strand, etc.).
        :rtype: coral.DNA

        '''
        copy = self.copy()
        copy._sequence, copy._bottom = copy._bottom, copy._sequence
        return copy

    def gc(self):
        '''Find the frequency of G and C in the current sequence.'''

        gc_n = len([base for base in self if str(base) == 'C' or
                    str(base) == 'G'])
        return float(gc_n) / len(self)

    def insert(self, sequence, index):
        inserted = super(DNA, self).insert(sequence, index)
        inserted.topology = self.topology
        return inserted

    def is_rotation(self, other):
        if len(self) != len(other):
            return False
        for i in range(len(self)):
            if self.rotate(i) == other:
                return True
        # If all else fails, check reverse complement
        rc = self.reverse_complement()
        for i in range(len(self)):
            if rc.rotate(i) == other:
                return True
        return False

    def linearize(self, index=0):
        '''Linearize circular DNA at an index.

        :param index: index at which to linearize.
        :type index: int
        :returns: A linearized version of the current sequence.
        :rtype: coral.DNA
        :raises: ValueError if the input is linear DNA.

        '''
        if self.topology == 'linear':
            raise ValueError('Cannot relinearize linear DNA.')
        copy = self.copy()
        copy.topology = 'linear'
        copy = copy[index:] + copy[:index]
        return copy

    def locate(self, pattern):
        '''Find sequences matching a pattern.

        :param pattern: Sequence for which to find matches.
        :type pattern: str
        :returns: A list of top and bottom strand indices of matches.
        :rtype: list of lists of indices (ints)
        :raises: ValueError if the pattern is longer than either the input
                 sequence (for linear DNA) or twice as long as the input
                 sequence (for circular DNA).

        '''
        # TODO: If linear, should use the methods in BaseSequence
        if self.topology == 'circular':
            if len(pattern) > 2 * len(self):
                raise ValueError('Pattern too long.')
        else:
            if len(pattern) > len(self):
                raise ValueError('Pattern too long.')

        pattern = str(pattern).upper()
        regex = '(?=' + pattern + ')'

        if self.topology == 'circular':
            r = len(pattern) - 1
            l = len(self) - r + 1
            top = self._sequence[l:] + self._sequence + self._sequence[:r]
            bottom = self._bottom[l:] + self._bottom + self._bottom[:r]
        else:
            top = self._sequence
            bottom = self._bottom

        top_starts = [index.start() for index in re.finditer(regex, top)]
        bottom_starts = [index.start() for index in re.finditer(regex, bottom)]

        # Adjust indices if doing circular search
        if self.topology == 'circular' and len(pattern) > 1:
            top_starts = [start - r + 1 for start in top_starts]
            bottom_starts = [start - r + 1 for start in bottom_starts]

        return [top_starts, bottom_starts]

    def mw(self):
        '''Calculate the molecular weight.

        :returns: The molecular weight of the current sequence.
        :rtype: float

        '''
        counter = collections.Counter((self._sequence + self._bottom).lower())
        mw_a = counter['a'] * 313.2
        mw_t = counter['t'] * 304.2
        mw_g = counter['g'] * 289.2
        mw_c = counter['c'] * 329.2
        return mw_a + mw_t + mw_g + mw_c

    def rotate(self, index):
        '''Orient DNA to index (only applies to circular DNA).

        :param index: DNA position at which to re-zero the DNA.
        :type index: int
        :returns: The current sequence reoriented at `index`.
        :rtype: coral.DNA
        :raises: ValueError if applied to linear sequence or `index` is
                 negative.

        '''
        if self.topology == 'linear' and index != 0:
            raise ValueError('Cannot rotate linear DNA')
        if index < 0:
            raise ValueError('Rotation index must be positive')
        else:
            return (self[index:] + self[0:index]).circularize()

    def rotate_by_feature(self, featurename):
        '''Reorient the DNA based on a feature it contains (circular DNA only).

        :param featurename: A uniquely-named feature.
        :type featurename: str
        :returns: The current sequence reoriented at the start index of a
                  unique feature matching `featurename`.
        :rtype: coral.DNA
        :raises: ValueError if there is no feature of `featurename` or
                 more than one feature matches `featurename`.

        '''
        # REFACTOR: Parts are redundant with .extract()
        matched = []
        for feature in self.features:
            if feature.name == featurename:
                matched.append(feature.copy())
        count = len(matched)
        if count == 1:
            return self.rotate(matched[0].start)
        elif count > 1:
            raise ValueError('More than one feature has that name.')
        else:
            raise ValueError('No such feature in the sequence.')

    def reverse_complement(self):
        '''Reverse complement the DNA.

        :returns: A reverse-complemented instance of the current sequence.
        :rtype: coral.DNA

        '''
        # TODO: put into NucleotideSequence class
        copy = self.copy()
        # Note: if sequence is double-stranded, swapping strand is basically
        # (but not entirely) the same thing - gaps affect accuracy.
        copy._sequence = reverse_complement(copy._sequence, 'dna')
        copy._bottom = reverse_complement(copy._bottom, 'dna')

        # Fix features (invert)
        for feature in copy.features:
            # Swap strand
            if feature.strand == 1:
                feature.strand = 0
            else:
                feature.strand = 1
            # Swap start and stop
            feature.start, feature.stop = (feature.stop, feature.start)
            # Adjust start/stop to feature len
            feature.start = len(copy) - feature.start
            feature.stop = len(copy) - feature.stop

        return copy

    def tm(self, parameters='cloning'):
        '''Find the melting temperature.

        :param parameters: The tm method to use (cloning, santalucia98,
                       breslauer86)
        :type parameters: str

        '''
        return coral.analysis.tm(self, parameters=parameters)

    def to_ss(self):
        '''Produce single stranded version of the current sequence.

        :returns: The current sequence, converted to ssDNA.
        :rtype: coral.DNA

        '''
        copy = self.copy()

        # Do nothing if already single-stranded
        if self.stranded == 'ss':
            return copy

        copy._bottom = '-' * len(copy)
        for top, bottom in zip(copy.top(), reversed(copy.bottom())):
            if top == bottom == '-':
                raise ValueError('Coercing to single-stranded would ' +
                                 'introduce a double stranded break.')
        copy.stranded = 'ss'

        return copy

    def to_ds(self):
        '''Produce double stranded version of the current sequence.

        :returns: The current sequence, converted to dsDNA.
        :rtype: coral.DNA

        '''
        # TODO: protect .stranded attribute if requiring setter method
        copy = self.copy()
        # Do nothing if already set
        if self.stranded == 'ds':
            return copy

        # Find strand that's all gaps (if ss this should be the case)
        reverse_seq = self.reverse_complement()
        if all([char == '-' for char in self._sequence]):
            copy._sequence = reverse_seq._bottom
        elif all([char == '-' for char in self._bottom]):
            copy._bottom = reverse_seq._sequence
        copy.stranded = 'ds'

        return copy

    def top(self):
        '''Return the raw string of the Watson (top) strand.

        :returns: The Watson strand.
        :rtype: str

        '''
        return self._sequence

    def transcribe(self):
        '''Transcribe into RNA.

        :returns: An RNA sequence transcribed from the current DNA sequence.
        :rtype: coral.RNA

        '''
        return coral.reaction.transcribe(self)

    def __add__(self, other):
        '''Add DNA together.

        :param other: instance to be added to.
        :type other: compatible sequence object (currently only DNA).
        :returns: Concatenated DNA sequence.
        :rtype: coral.DNA
        :raises: Exception if either sequence is circular.
                 Exception if concatenating a sequence with overhangs would
                 create a discontinuity.

        '''
        if type(self) != type(other):
            try:
                other = type(self)(other)
            except AttributeError:
                raise TypeError('Cannot add {} to {}'.format(self, other))

        if self.topology == 'circular' or other.topology == 'circular':
            raise Exception('Can only add linear DNA.')

        discontinuity = [False, False]
        if len(self) != 0 and len(other) != 0:
            # If either is empty, let things proceed anyways
            discontinuity[0] = (self._sequence[-1] == '-' and
                                other._bottom[-1] == '-')
            discontinuity[1] = (self._bottom[0] == '-' and
                                other._sequence[0] == '-')

        for_discontinuity = discontinuity[0]
        rev_discontinuity = discontinuity[1]

        if for_discontinuity or rev_discontinuity:
            msg = 'Concatenated DNA would be discontinuous.'
            raise Exception(msg)

        if self.stranded == 'ds' or other.stranded == 'ds':
            stranded = 'ds'
        else:
            stranded = 'ss'

        tops = self._sequence + other._sequence
        bottoms = other._bottom + self._bottom
        self_features = [feature.copy() for feature in self.features]
        other_features = [feature.copy() for feature in other.features]
        for feature in other_features:
            feature.move(len(self))
        features = self_features + other_features

        new_instance = DNA(tops, bottom=bottoms, topology='linear',
                           stranded=stranded, run_checks=False,
                           features=features)

        return new_instance

    def __contains__(self, query):
        '''Defines `query in sequence` operator.

        :param query: query string or DNA sequence
        :type query: str or coral.DNA

        '''
        # query in forward sequence
        if super(DNA, self).__contains__(query, 'N'):
            return True
        # query in reverse complement
        elif super(DNA, self.reverse_complement()).__contains__(query, 'N'):
            return True
        # query in neither
        else:
            return False

    def __delitem__(self, index):
        '''Delete sequence at an index.

        :param index: index to delete
        :type index: int
        :returns: The current sequence with the base at `index` removed.
        :rtype: coral.DNA

        '''
        super(DNA, self).__delitem__(index)
        bottom_list = list(self._bottom[::-1])
        del bottom_list[index]
        self._bottom = ''.join(bottom_list)[::-1]

    def __getitem__(self, key):
        '''Index and slice sequences.

        :param key: int or slice object for subsetting.
        :type key: int or slice object
        :returns: A subsequence matching the slice (`key`).
        :rtype: coral.DNA

        '''
        # Use BaseSequence method to assign top strand and figure out features
        if isinstance(key, slice):
            if all([k is None for k in [key.start, key.stop, key.step]]):
                # It's the copy slice operator ([:])
                return self.copy()
            else:
                # The key is a normal slice
                copy = super(DNA, self).__getitem__(key)

                # bottom_key = slice(-key.stop if key.stop is not None
                #                    else None,
                #                    -key.start if key.start is not None
                #                    else None,
                #                    key.step)
                # copy._bottom = copy._bottom[bottom_key]
                copy._bottom = copy._bottom[::-1][key][::-1]
        else:
            # The key is an integer
            copy = super(DNA, self).__getitem__(key)
            copy._bottom = copy._bottom[-key]

        copy.topology = 'linear'

        return copy

    def __eq__(self, other):
        '''Define equality - sequences, topology, and strandedness are the
        same.

        :returns: Whether current sequence's (Watson and Crick), topology,
                  and strandedness are equivalent to those of another sequence.
        :rtype: bool

        '''
        tops_equal = self._sequence == other._sequence
        bottoms_equal = self._bottom == other._bottom
        topology_equal = self.topology == other.topology
        stranded_equal = self.stranded == other.stranded
        if tops_equal and bottoms_equal and topology_equal and stranded_equal:
            return True
        else:
            return False

    def __repr__(self):
        '''String to print when object is called directly.'''
        parent = super(DNA, self).__repr__()
        display_bases = 40
        if len(self._sequence) < 90:
            bottom = self._bottom[::-1]
        else:
            rev_bottom = self._bottom[::-1]
            bottom = ''.join([rev_bottom[0:display_bases], ' ... ',
                              rev_bottom[-display_bases:]])
        first_line = '{} {}DNA:'.format(self.topology, self.stranded)
        to_print = '\n'.join([first_line, parent, bottom])
        return to_print

    def __setitem__(self, index, new_value):
        '''Sets value at index to new value.

        :param index: The index at which the sequence will be modified.
        :type index: int
        :param new_value: The new value at that index
        :type new_value: str or coral.DNA
        :returns: The current sequence with the sequence at `index` replaced
                  with `new_value`.
        :rtype: coral.DNA
        :raises: ValueError if `new_value` is '-'.

        '''
        new_value = str(new_value)
        if new_value == '-':
            raise ValueError('Cannot insert gap - split sequence instead.')
        # setitem on top strand
        super(DNA, self).__setitem__(index, new_value)
        # setitem on bottom strand
        if self.stranded == 'ds':
            sequence_list = list(self._bottom)[::-1]
            sequence_list[index] = str(DNA(new_value).reverse_complement())
            self._bottom = ''.join(sequence_list[::-1])
        else:
            self._bottom = '-' * len(self)


class RestrictionSite(object):
    '''Recognition site and properties of a restriction endonuclease.'''
    def __init__(self, recognition_site, cut_site, name=None):
        '''
        :param recognition_site: Input sequence.
        :type recognition_site: coral.DNA
        :param cut_site: 0-indexed indices where DNA is nicked (top, then
                         bottom strand). For an n-sized recognition site, there
                         are n + 1 positions at which to cut.
        :type cut_site: 2-tuple.
        :param name: Identifier of this restriction site
        :type name: str
        :returns: instance of coral.RestrictionSite

        '''
        self.recognition_site = recognition_site  # require DNA object
        # cutsite is indexed to leftmost base of restriction site
        self.cut_site = cut_site  # tuple of where top/bottom strands are cut
        # optional name
        self.name = name

    def is_palindrome(self):
        '''Report whether sequence is palindromic.

        :returns: Whether the restriction site is a palindrome.
        :rtype: bool

        '''
        return self.recognition_site.is_palindrome()

    def cuts_outside(self):
        '''Report whether the enzyme cuts outside its recognition site.
        Cutting at the very end of the site returns True.

        :returns: Whether the enzyme will cut outside its recognition site.
        :rtype: bool

        '''
        for index in self.cut_site:
            if index < 0 or index > len(self.recognition_site) + 1:
                return True
        return False

    def copy(self):
        '''Return copy of the restriction site.

        :returns: A safely editable copy of the current restriction site.
        :rtype: coral.RestrictionSite

        '''
        return RestrictionSite(self.recognition_site, self.cut_site,
                               self.name)

    def __repr__(self):
        '''Represent a restriction site.'''
        site = self.recognition_site
        cut_symbols = ('|', '|')
        if not self.cuts_outside():
            top_left = str(site[0:self.cut_site[0]])
            top_right = str(site[self.cut_site[0]:])
            top_w_cut = top_left + cut_symbols[0] + top_right

            bottom_left = site[0:self.cut_site[1]].reverse_complement()
            bottom_left = str(bottom_left)[::-1]
            bottom_right = site[self.cut_site[1]:].reverse_complement()
            bottom_right = str(bottom_right)[::-1]
            bottom_w_cut = bottom_left + cut_symbols[1] + bottom_right
        else:
            return '\n'.join([site.top() + ' {}'.format(self.cut_site),
                              site.bottom()])

        return '\n'.join([top_w_cut, bottom_w_cut])

    def __len__(self):
        '''Defines len operator.

        :returns: Length of the recognition site.
        :rtype: int

        '''
        return len(self.recognition_site)


class Primer(object):
    '''A DNA primer - ssDNA with tm, anneal, and optional overhang.'''
    def __init__(self, anneal, tm, overhang=None, name='', note=''):
        '''
        :param anneal: Annealing sequence
        :type anneal: coral.DNA
        :param overhang: Overhang sequence
        :type overhang: coral.DNA
        :param tm: melting temperature
        :type tm: float
        :param name: Optional name of the primer. Used when writing to csv with
                     seqio.write_primers.
        :type name: str
        :param note: Optional description to associate with the primer. Used
                     when writing to csv with seqio.write_primers.
        :type note: str
        :returns: coral.Primer instance.

        '''
        self.tm = tm
        self.anneal = anneal.to_ss()
        if overhang is not None:
            self.overhang = overhang.to_ss()
        else:
            self.overhang = DNA('', stranded='ss')
        self.name = name
        self.note = note

    def copy(self):
        '''Generate a Primer copy.

        :returns: A safely-editable copy of the current primer.
        :rtype: coral.DNA

        '''
        return type(self)(self.anneal, self.tm, overhang=self.overhang,
                          name=self.name, note=self.note)

    def primer(self):
        '''Produce full (overhang + annealing sequence) primer sequence.

        :returns: The DNA sequence of the primer.
        :rtype: coral.DNA

        '''
        return self.overhang + self.anneal

    def __repr__(self):
        '''Representation of a primer.'''
        if self.overhang:
            return 'Primer: {} Tm: {:.2f}'.format(self.overhang.top().lower() +
                                                  self.anneal.top(), self.tm)
        else:
            return 'Primer: {} Tm: {:.2f}'.format(self.anneal.top(), self.tm)

    def __str__(self):
        '''Coerce DNA object to string.

        :returns: A string of the full primer sequence.
        :rtype: str

        '''
        return str(self.primer())

    def __eq__(self, other):
        '''Define equality - sequences, topology, and strandedness are the
        same.

        :returns: Whether two primers have the same overhang and annealing
                  sequence.
        :rtype: bool

        '''
        anneal_equal = self.anneal == other.anneal
        overhang_equal = self.overhang == other.overhang
        if anneal_equal and overhang_equal:
            return True
        else:
            return False

    def __len__(self):
        '''Define len operator.

        :returns: The length of the full primer sequence.
        :rtype: int

        '''
        return len(self.primer())



'''Coral, core modules for synthetic DNA design.'''
try:
    import matplotlib
except ImportError:
    print "Failed to import matplotlib. Plotting sequencing won't work."

from . import analysis
from . import constants
from . import database
from . import design
from . import reaction
from . import seqio
from .sequence import DNA
from .sequence import RNA
from .sequence import Peptide
from .sequence import Primer
from .sequence import RestrictionSite
from .sequence import Feature



'''Base sequence classes.'''
import re
from coral.constants.genbank import TO_CORAL
from coral.constants.molecular_bio import ALPHABETS, COMPLEMENTS


class BaseSequence(object):
    '''Base sequence class.'''
    def __init__(self, sequence, material, features=None, run_checks=True):
        '''
        :param sequence: Input sequence.
        :type sequence: str
        :param material: Material type (dna, rna, peptide)
        :type material: str
        :param features: List of annotated features.
        :type features: list
        :param run_checks: Check inputs / formats (disabling increases speed):
                           alphabet check
                           case
        :type run_checks: bool
        :returns: coral.sequence.BaseSequence instance.

        '''
        if run_checks:
            self._sequence = process_seq(sequence, material)
        else:
            self._sequence = sequence

        self._material = material

        if features is None:
            self.features = []
        else:
            self.features = features

    def annotate_from_library(self, library, wipe=True, shortest=6):
        '''Annotate the sequence using the features of another. Ignores
        features shorter than 5 bp.

        :param library: list of features with a .sequence attribute
        :type library: list of features with a .sequence attribute
        :param wipe: Remove (wipe) the current features first.
        :type wipe: bool
        :param shortest: Features shorters than this will be ignored.
        :type shortest: int

        '''
        copy = self.copy()
        if wipe:
            copy.features = []

        # Make sure features are unique and above 'shortest' parameter
        unique_features = []
        unique_sequences = []
        for feature in library:
            if len(feature.sequence) >= shortest:
                if feature.sequence not in unique_sequences:
                    unique_features.append(feature)
                    unique_sequences.append(feature.sequence)
        # Match features
        for feature in unique_features:
            if feature.sequence in copy:
                match_location = copy.locate(feature.sequence)
                # Only annotate the top strand if sequence is palindrome
                if feature.sequence.is_palindrome():
                    match_location[1] = []
                # Which strand is it on?
                for i, strand in enumerate(match_location):
                    for match in strand:
                        new_feature = feature.copy()
                        length = new_feature.stop - new_feature.start
                        if i == 0:
                            # Watson strand
                            new_feature.start = match
                            new_feature.stop = new_feature.start + length
                        else:
                            # Crick strand
                            new_feature.stop = len(copy) - match
                            new_feature.start = new_feature.stop - length
                            # If feature found on other strand, update
                            new_feature.strand = abs(new_feature.strand - 1)
                        # Modulo just in case it spans the origin
                        new_feature.start = new_feature.start % len(copy)
                        new_feature.stop = new_feature.stop % len(copy)
                        copy.features.append(new_feature)
        return copy

    def annotate_from_other(self, other, wipe=True, shortest=6):
        '''Annotate the sequence using the features of another. Ignores
        features shorter than 5.

        :param other: Another sequence.
        :type other: coral.DNA
        :param wipe: Remove (wipe) the current features first.
        :type wipe: bool
        :param shortest: Features shorters than this will be ignored.
        :type shortest: int

        '''
        # Generate feature library
        features = [feature.copy() for feature in other.features]
        for feature in features:
            feature.sequence = other[feature.start:feature.stop]

        # Annotate from library
        return self.annotate_from_library(features, wipe=wipe,
                                          shortest=shortest)

    def copy(self):
        '''Create a copy of the current instance.

        :returns: A safely editable copy of the current sequence.

        '''
        # Significant performance improvements by skipping alphabet check
        if self.features:
            features = [feature.copy() for feature in self.features]
        else:
            features = []
        return type(self)(self._sequence, self._material,
                          features=features, run_checks=False)

    def endswith(self, seq):
        '''Report whether parent sequence ends with a query sequence.

        :param seq: Query sequence.
        :returns: Boolean of whether the sequence ends with the query.
        :rtype: bool

        '''
        if self._sequence.endswith(str(seq)):
            return True
        else:
            return False

    def extract(self, feature, any_char, remove_subfeatures=False):
        '''Extract a feature from the sequence.

        :param feature: Feature object.
        :type feature: coral.sequence.Feature
        :param any_char: Turn any gaps in the feature into Ns or Xs and remove
                         all other features. If False, just extracts start:stop
                         slice.
        :type any_char: bool
        :param remove_subfeatures: Remove all features in the extracted
                                   sequence aside from the input feature.
        :type remove_subfeatures: bool
        :returns: A subsequence from start to stop of the feature.

        '''
        extracted = self[feature.start:feature.stop]
        # Turn gaps into Ns or Xs
        for gap in feature.gaps:
            for i in range(*gap):
                extracted[i] = any_char
        if remove_subfeatures:
            # Keep only the feature specified
            extracted.features = [feature]
        # Update feature locations
        # copy them
        for feature in extracted.features:
            feature.move(-feature.start)
        return extracted

    def insert(self, sequence, index):
        '''Insert a sequence at index.

        :param sequence: Sequence to insert
        :param index: (internal) index at which to insert the sequence.
        :type index: int

        '''
        range_error = IndexError('Invalid index - must be between 1 and ' +
                                 'length - 1.')
        if index == 0:
            raise range_error
        try:
            self[index]
        except IndexError:
            raise range_error

        return self[0:index] + sequence + self[index:]

    def locate(self, pattern):
        '''Find sequences matching a pattern.

        :param pattern: Sequence for which to find matches.
        :type pattern: str
        :returns: Indices of pattern matches.
        :rtype: list of ints

        '''
        if len(pattern) > len(self):
            raise ValueError('Pattern too long.')
        pattern = str(pattern).upper()
        re_pattern = '(?=' + pattern + ')'
        return [index.start() for index in
                re.finditer(re_pattern, self._sequence)]

    def startswith(self, query):
        '''Report whether parent sequence starts with a query sequence.

        :param seq: Query sequence.
        :type seq: str or coral.DNA
        :returns: Boolean of whether the top strand starts with the query.
        :rtype: bool

        '''
        if self._sequence.startswith(str(query)):
            return True
        else:
            return False

    def __add__(self, other):
        '''Defines addition.

        :param other: Instance with which to sum.
        :type other: coral.sequence.BaseSequence
        :returns: Concatenated sequence.
        :rtype: coral.sequence.BaseSequence

        '''
        if type(self) != type(other):
            try:
                other = type(self)(other)
            except AttributeError:
                raise TypeError('Cannot add {} to {}'.format(self, other))

        self_features = [feature.copy() for feature in self.features]
        other_features = [feature.copy() for feature in other.features]
        for feature in other_features:
            feature.move(len(self))
        features = self_features + other_features

        return BaseSequence(self._sequence + other._sequence, self._material,
                            features=features, run_checks=False)

    def __contains__(self, query, any_char):
        '''`x in y`.

        :param query: Query (i.e. exact pattern) sequence to find.
        :type query: str
        :param any_char: Character to use for any match (.*).
        :type any_char: str
        :returns: Whether the query is found in the current sequence.
        :rtype: bool

        '''
        query_str = str(query).upper()
        query_str = re.sub(any_char, '.', query_str)
        if re.search(query_str, str(self)):
            return True
        else:
            return False

    def __delitem__(self, index):
        '''Deletes sequence at index.

        :param index: Index to delete
        :type index: int
        :returns: The current sequence with the moiety at `index` removed.
        :rtype: coral.sequence.BaseSequence

        '''
        if self.features:
            self.features = [feature for feature in self.features if index not
                             in range(feature.start, feature.stop)]
            for feature in self.features:
                if feature.start >= index:
                    feature.move(-1)

        sequence_list = list(self._sequence)
        del sequence_list[index]
        self._sequence = ''.join(sequence_list)

    def __eq__(self, other):
        '''Define == operator. True if sequences are the same.

        :param other: Other sequence.
        :type other: coral.sequence._sequence.BaseSequence
        :returns: Whether two sequences have the same base string (sequence).
        :rtype: bool

        '''
        if self._sequence == other._sequence:
            return True
        else:
            return False

    def __getitem__(self, key):
        '''Indexing and slicing of sequences.

        :param key: int or slice for subsetting.
        :type key: int or slice
        :returns: Slice of the current sequence.
        :rtype: coral.sequence.BaseSequence

        '''
        copy = self.copy()
        copy._sequence = self._sequence[key]

        # If empty sequence, return empty sequence
        if not copy._sequence:
            copy.features = []
            return copy

        # If sequence has features, update them
        def in_slice(feature):
            if key.start is not None and feature.start < key.start:
                return False
            elif key.stop is not None and feature.stop > key.stop:
                return False
            else:
                return True

        if copy.features:
            if isinstance(key, slice):
                # If a slice, remove stuff that isn't in the slide and
                # adjust feature starts/stops
                if key.step == 1 or key.step is None:
                    copy.features = [feature.copy() for feature in
                                     self.features if in_slice(feature)]
                    if key.start:
                        for feature in copy.features:
                            feature.move(-key.start)
                else:
                    copy.features = []
            else:
                copy.features = []
                for feature in self.features:
                    if feature.start == feature.stop == key:
                        copy.features.append(feature.copy())
                        copy.features[-1].move(key)
        return copy

    def __len__(self):
        '''Calculate sequence length.

        :returns: The length of the sequence.
        :rtype: int

        '''
        return len(self._sequence)

    def __mul__(self, n):
        '''Concatenate copies of the sequence.

        :param n: Factor by which to multiply the sequence.
        :type n: int
        :returns: The current sequence repeated n times.
        :rtype: coral.sequence.BaseSequence
        :raises: TypeError if n is not an integer.

        '''
        # Input checking
        if n != int(n):
            raise TypeError('Multiplication by non-integer.')
        return sum([x for x in _decompose(self, n)])

    def __ne__(self, other):
        '''Define != operator.

        :param other: Other sequence.
        :type other: coral.sequence._sequence.BaseSequence
        :returns: The opposite of ==.
        :rtype: bool

        '''
        try:
            return not (self == other)
        except TypeError:
            return False

    def __radd__(self, other):
        '''Add unlike types (enables sum function).

        :param other: Object of any other type.
        :param other: coral.sequence._sequence.BaseSequence
        :returns: Concatenated sequence.
        :rtype: coral.sequence.BaseSequence

        '''
        if other == 0 or other is None:
            # For compatibility with sum()
            return self
        elif type(self) != type(other):
            try:
                other = type(self)(other)
            except AttributeError:
                raise TypeError('Cannot add {} to {}'.format(self, other))
        return self + other

    def __repr__(self):
        '''String to print when object is called directly.'''
        display_bases = 40
        if len(self._sequence) < 90:
            sequence = self._sequence
        else:
            sequence = ''.join([self._sequence[:display_bases], ' ... ',
                                self._sequence[-display_bases:]])
        return sequence

    def __setitem__(self, index, new_value):
        '''Sets index value to new value.

        :param index: Index to modify.
        :type index: int
        :param new_value: Value to input.
        :type new_value: str or coral.sequence._sequence.BaseSequence
        :returns: The current sequence with the moiety at `index` replaced
                  by `new_value`.
        :rtype: coral.sequence.BaseSequence

        '''
        if self.features:
            for i, feature in enumerate(self.features[::-1]):
                if index in range(feature.start, feature.stop):
                    self.features.pop(-(i + 1))

        sequence_list = list(self._sequence)
        sequence_list[index] = str(BaseSequence(new_value, self._material))
        self._sequence = ''.join(sequence_list)

    def __str__(self):
        '''Cast to string.

        :returns: A string of the current sequence
        :rtype: str

        '''
        return self._sequence


class NucleotideSequence(BaseSequence):
    '''Nucleotide sequence class.'''
    def is_palindrome(self):
        '''Report whether sequence is palindromic.

        :returns: Boolean stating whether sequence is a palindrome.
        :rtype: bool

        '''
        return palindrome(self)


def _decompose(string, n):
    '''Given string and multiplier n, find m**2 decomposition.

    :param string: input string
    :type string: str
    :param n: multiplier
    :type n: int
    :returns: generator that produces m**2 * string if m**2 is a factor of n
    :rtype: generator of 0 or 1

    '''
    binary = [int(x) for x in bin(n)[2:]]
    new_string = string
    counter = 1
    while counter <= len(binary):
        if binary[-counter]:
            yield new_string
        new_string += new_string
        counter += 1


class Feature(object):
    '''Represent A DNA feature - annotate and extract sequence by metadata.'''
    def __init__(self, name, start, stop, feature_type, gene='', locus_tag='',
                 qualifiers={}, strand=0, gaps=[]):
        '''
        :param name: Name of the feature. Used during feature extraction.
        :type name: str
        :param start: Where the feature starts
        :type start: int
        :param stop: Where the feature stops
        :type stop: int
        :param feature_type: The type of the feature. Allowed types:
                                'coding', 'primer', 'promoter', 'terminator',
                                'rbs'
        :type name: str
        :param strand: Watson (0) or Crick (1) strand of the feature.
        :type strand: int
        :param gaps: Gap locations if the feature has gaps.
        :type gaps: list of coordinates (2-tuple/list)
        :param gene: gene attribute (Genbank standard) - gene name (e.g. galK
                     on MG1655 genome).
        :type gene: str
        :param locus_tag: locus_tag attribute (Genbank standard) - systematic
                          locus name (e.g. b0757 for galK on MG1655 genome).
        :type locus_tag: str
        :param qualifiers: Complete Genbank qualifiers key:value pairs
        :type qualifiers: dict
        :returns: coral.Feature instance.
        :raises: ValueError if `feature_type` is not in
                 coral.constants.genbank.TO_CORAL.

        '''
        self.name = name
        self.start = int(start)
        self.stop = int(stop)
        self.modified = False
        self.gene = gene
        self.locus_tag = locus_tag
        self.qualifiers = qualifiers
        self.strand = strand
        self.gaps = gaps

        allowed_types = TO_CORAL.keys()

        if feature_type in allowed_types:
            self.feature_type = feature_type
        else:
            msg1 = 'feature_type'
            msg2 = 'must be one of the following: {}'.format(allowed_types)
            raise ValueError(msg1 + msg2)

    def move(self, bases):
        '''Move the start and stop positions.

        :param bases: bases to move - can be negative
        :type bases: int

        '''
        self.start += bases
        self.stop += bases

    def copy(self):
        '''Return a copy of the Feature.

        :returns: A safely editable copy of the current feature.
        :rtype: coral.Feature

        '''
        return type(self)(self.name, self.start, self.stop, self.feature_type,
                          gene=self.gene, locus_tag=self.locus_tag,
                          qualifiers=self.qualifiers, strand=self.strand)

    def __repr__(self):
        '''Represent a feature.'''
        if self.modified:
            part1 = '(Modified) {} "{}" feature '.format(self.name,
                                                         self.feature_type)
        else:
            part1 = '{} "{}" feature '.format(self.name, self.feature_type)
        part2 = '({0} to {1}) on strand {2}'.format(self.start, self.stop,
                                                    self.strand)
        return part1 + part2

    def __eq__(self, other):
        '''Define equality.

        :returns: Whether the name and feature type are the same.
        :rtype: bool

        '''
        if self.name != other.name:
            return False
        if self.feature_type != other.feature_type:
            return False
        if self.start != other.start:
            return False
        if self.stop != other.stop:
            return False
        if self.strand != other.strand:
            return False
        if self.gaps != other.gaps:
            return False

        return True

    def __ne__(self, other):
        '''Define inequality.'''
        if not self == other:
            return True
        else:
            return False


def reverse_complement(sequence, material):
    '''Reverse complement a sequence.

    :param sequence: Sequence to reverse complement
    :type sequence: str
    :param material: dna, rna, or peptide.
    :type material: str
    '''
    code = dict(COMPLEMENTS[material])
    reverse_sequence = sequence[::-1]
    return ''.join([code[base] for base in reverse_sequence])


def check_alphabet(seq, material):
    '''Verify that a given string is valid DNA, RNA, or peptide characters.

    :param seq: DNA, RNA, or peptide sequence.
    :type seq: str
    :param material: Input material - 'dna', 'rna', or 'pepide'.
    :type sequence: str
    :returns: Whether the `seq` is a valid string of `material`.
    :rtype: bool
    :raises: ValueError if `material` isn't "dna", "rna", or "peptide".
             ValueError if `seq` contains invalid characters for its
             material type.

    '''
    errs = {'dna': 'DNA', 'rna': 'RNA', 'peptide': 'peptide'}
    if material == 'dna' or material == 'rna' or material == 'peptide':
        alphabet = ALPHABETS[material]
        err_msg = errs[material]
    else:
        msg = 'Input material must be "dna", "rna", or "peptide".'
        raise ValueError(msg)
    # This is a bottleneck when modifying sequence - hence the run_checks
    # optional parameter in sequence objects..
    # First attempt with cython was slower. Could also try pypy.
    if re.search('[^' + alphabet + ']', seq):
        raise ValueError('Encountered a non-%s character' % err_msg)


def process_seq(seq, material):
    '''Validate and process sequence inputs.

    :param seq: input sequence
    :type seq: str
    :param material: DNA, RNA, or peptide
    :type: str
    :returns: Uppercase version of `seq` with the alphabet checked by
              check_alphabet().
    :rtype: str

    '''
    check_alphabet(seq, material)
    seq = seq.upper()
    return seq


def palindrome(seq):
    '''Test whether a sequence is palindrome.

    :param seq: Sequence to analyze (DNA or RNA).
    :type seq: coral.DNA or coral.RNA
    :returns: Whether a sequence is a palindrome.
    :rtype: bool

    '''
    seq_len = len(seq)
    if seq_len % 2 == 0:
        # Sequence has even number of bases, can test non-overlapping seqs
        wing = seq_len / 2
        l_wing = seq[0: wing]
        r_wing = seq[wing:]
        if l_wing == r_wing.reverse_complement():
            return True
        else:
            return False
    else:
        # Sequence has odd number of bases and cannot be a palindrome
        return False



'''RNA module.'''
import coral.reaction
from ._sequence import NucleotideSequence
from ._sequence import reverse_complement


class RNA(NucleotideSequence):
    '''RNA sequence.'''
    def __init__(self, rna, run_checks=True):
        '''
        :param rna: Input sequence (RNA).
        :type rna: str
        :param run_checks: Check inputs / formats (disabling increases speed):
                           alphabet check
                           case
        :type run_checks: bool
        :returns: coral.RNA instance.

        '''
        super(RNA, self).__init__(rna, 'rna', run_checks=run_checks)

    def copy(self):
        '''Create a copy of the current instance.

        :returns: A safely-editable copy of the current sequence.
        :rtype: coral.RNA

        '''
        # Significant performance improvements by skipping alphabet check
        return type(self)(self._sequence, run_checks=False)

    def reverse_complement(self):
        '''Reverse complement sequence.

        :returns: The reverse-complement of the current sequence.
        :rtype: coral.RNA

        '''
        new_instance = self.copy()
        new_instance._sequence = reverse_complement(self._sequence, 'rna')
        return new_instance

    def extract(self, name, remove_subfeatures=False):
        return super(RNA, self).extract(self, name, 'N',
                                        remove_subfeatures=remove_subfeatures)

    def reverse_transcribe(self):
        '''Reverse transcribe to DNA.

        :returns: The reverse transcribed (DNA) version of the current RNA.
        :rtype: coral.DNA

        '''
        return coral.reaction.reverse_transcribe(self)

    def translate(self):
        '''Translate sequence into a peptide.

        :returns: A translated peptide from the current sequence.
        :rtype: coral.Peptide

        '''
        return coral.reaction.translate(self)

    def __contains__(self, query):
        '''Defines `query in sequence` operator.

        :param query: query string or DNA sequence
        :type query: str or coral.DNA

        '''
        return super(RNA, self).__contains__(query, 'N')



'''Peptide module.'''
from ._sequence import BaseSequence


class Peptide(BaseSequence):
    '''Peptide sequence.'''
    def __init__(self, peptide, run_checks=True):
        '''
        :param peptide: Input sequence (peptide).
        :type peptide: str
        :param run_checks: Check inputs / formats (disabling increases speed):
                           alphabet check
                           case
        :type run_checks: bool
        :returns: coral.Peptide instance.

        '''
        super(Peptide, self).__init__(peptide, 'peptide',
                                      run_checks=run_checks)

    def copy(self):
        '''Create a copy of the current instance.

        :returns: A safely editable copy of the current sequence.
        :rtype: coral.Peptide

        '''
        # Significant performance improvements by skipping alphabet check
        return type(self)(self._sequence, run_checks=False)

    def extract(self, name, pure=False):
        return super(Peptide, self).extract(self, name, 'X', pure=True)

    def __contains__(self, query):
        '''Defines `query in sequence` operator.

        :param query: query string or DNA sequence
        :type query: str or coral.DNA

        '''
        return super(Peptide, self).__contains__(query, 'X')



"""Resection (need a new name!) - exonuclease activity."""


def five_resect(dna, n_bases):
    """Remove bases from 5' end of top strand.

    :param dna: Sequence to resect.
    :type dna: coral.DNA
    :param n_bases: Number of bases cut back.
    :type n_bases: int
    :returns: DNA sequence resected at the 5' end by n_bases.
    :rtype: coral.DNA

    """
    new_instance = dna.copy()
    new_top = '-' * min(len(dna), n_bases) + str(dna)[n_bases:]
    new_instance._sequence = new_top
    new_instance = _remove_end_gaps(new_instance)
    if n_bases >= len(dna):
        new_instance._sequence = ''.join(['-' for i in range(len(dna))])
        new_instance.stranded = 'ss'
    return new_instance


def three_resect(dna, n_bases):
    """Remove bases from 3' end of top strand.

    :param dna: Sequence to resect.
    :type dna: coral.DNA
    :param n_bases: Number of bases cut back.
    :type n_bases: int
    :returns: DNA sequence resected at the 3' end by n_bases.
    :rtype: coral.DNA

    """
    new_instance = dna.copy()

    new_top = str(dna)[:-n_bases] + '-' * min(len(dna), n_bases)
    new_instance._sequence = new_top
    new_instance = _remove_end_gaps(new_instance)
    if n_bases >= len(dna):
        new_instance._sequence = ''.join(['-' for i in range(len(dna))])
        new_instance.stranded = 'ss'
    return new_instance


def _remove_end_gaps(sequence):
    """Removes double-stranded gaps from ends of the sequence.

    :returns: The current sequence wiht terminal double-strand gaps ('-')
              removed.
    :rtype: coral.DNA

    """
    top = sequence.top()
    bottom_rev = sequence.bottom()[::-1]

    top_lstrip = top.lstrip('-')
    bottom_lstrip = bottom_rev.lstrip('-')
    lstrip_len = max(len(top_lstrip), len(bottom_lstrip))
    top = top[-lstrip_len:]
    bottom_rev = bottom_rev[-lstrip_len:]

    top_rstrip = top.rstrip('-')
    bottom_rstrip = bottom_rev.rstrip('-')
    rstrip_len = max(len(top_rstrip), len(bottom_rstrip))
    top = top[0:rstrip_len]
    bottom_rev = bottom_rev[0:rstrip_len]

    bottom = bottom_rev[::-1]
    sequence._sequence = top
    sequence._bottom = bottom
    return sequence



try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension

import numpy
# For rebuilding Cython extensions
try:
    from Cython.Distutils import build_ext
    USECYTHON = True
except ImportError:
    USECYTHON = False


config = {
    'description': 'coral',
    'author': 'Nick Bolten',
    'url': 'https://github.com/klavinslab/coral',
    'download_url': 'https://github.com/klavinslab/coral.git',
    'author_email': 'nbolten _at_ gmail',
    'version': '0.1.0',
    'install_requires': ['nose', 'numpy', 'biopython', 'intermine',
                         'requests'],
    'extras_require': {'plotting': ['matplotlib'],
                       'documentation': ['sphinx']},
    'packages': ['coral',
                 'coral.analysis',
                 'coral.analysis._sequence',
                 'coral.analysis._sequencing',
                 'coral.analysis._structure',
                 'coral.constants',
                 'coral.database',
                 'coral.design',
                 'coral.design._oligo_synthesis',
                 'coral.design._sequence_generation',
                 'coral.seqio',
                 'coral.reaction',
                 'coral.sequence'],
    'package_data': {'coral': ['coral/analysis/_sequencing/data/*']},
    'include_package_data': True,
    'scripts': [],
    'name': 'coral',
    'license': 'Copyright University of Washington'
}

seq_extension = Extension('coral.analysis._sequencing.calign',
                          ['coral/analysis/_sequencing/calign.c'],
                          include_dirs=[numpy.get_include()])
EXTENSIONS = [seq_extension]

if USECYTHON:
    setup(cmdclass={'build_ext': build_ext},
          ext_modules=EXTENSIONS,
          test_suite='nose.collector',
          **config)
else:
    setup(ext_modules=EXTENSIONS,
          test_suite='nose.collector',
          **config)
